# Copyright (c) 2013, University of Bristol and University of Southampton.

import scipy
import numpy
import numpy.ma
import time
import os
import subprocess
import logging
import re

from mako.lookup import TemplateLookup
from mako.template import Template as MakoTemplate
from collections import OrderedDict

import uuid

import EStat
from EStat.Parameter import Parameter
from EStat.Parameter import RunningStat
from EStat.Parameter import RunningStatVector

from EStat.stats.utils import lock

from EStat.DTAFile import DTAFile
from EStat.Templating import *
from EStat.XMLtoSubC import XMLtoSubC

from EStat import bitbmsl
import ply.lex as lex
import ply.yacc as yacc
from xml.dom import minidom

from lxml import etree

from EStat.Templating import DataScalar, ParamScalar, DataVector, IDVector, ParamVector, DataMatrix, IDMatrix, ParamMatrix
from EStat.engines.Demo import Demo

from EStat.stats.utils.MCMCGraph import MCMCGraph
from EStat.stats.utils.MCMCChains import MCMCChains

from EStat.NumUtils import numpy_eval
from EStat.NumUtils import flatten_list

xsltpath = os.path.abspath(os.path.join(os.path.split(os.path.abspath(__file__))[0], '..', 'eStat', 'bin', 'Display'))

tmplpath = os.path.abspath(os.path.join(os.path.split(os.path.abspath(__file__))[0], '..', 'src', 'lib', 'EStat', 'templates'))
stdtmpl = TemplateLookup(directories=[tmplpath])

import threading

class ModelThread(threading.Thread):

    __version__ = '1.0.0'

    def __init__(self, callback, chainnum, current_iteration, total_iterations, callback_iterations, code, scode, locals, globals):
        threading.Thread.__init__(self)
        self.callback = callback
        self.current_iteration = current_iteration
        self.total_iterations = total_iterations
        self.callback_iterations = sorted(callback_iterations)
        self.chainnum = chainnum
        self.code = code
        self.scode = scode
        self.locals = locals
        self.globals = globals

    def run(self):
        while self.current_iteration != self.total_iterations:
            next_iteration = self.total_iterations
            while len(self.callback_iterations) > 0:
                impit = self.callback_iterations.pop(0)
                if impit > self.current_iteration:
                    next_iteration = impit
                    break

            self.locals['numiter'] = next_iteration - self.current_iteration

            #logging.info('calculating iterations ' + str(self.current_iteration) + ' to ' + str(next_iteration))

            res = scipy.weave.inline(self.code, 
                                  self.locals.keys() + self.globals.keys(),
                                  local_dict = self.locals, 
                                  global_dict = self.globals, 
                                  support_code = self.scode, 
                                  compiler = 'gcc', 
                                  headers = ['<unordered_map>', '<vector>', '<cmath>', '<cfloat>', '<limits>', '<random>', '"rng.h"', '"statlib.h"'],
                                  include_dirs = [tmplpath],
                                  sources = [os.path.join(tmplpath, 'statlib.cpp'), os.path.join(tmplpath, 'rng_simple.cpp')],
                                  library_dirs = [tmplpath],
                                  libraries = ['blas', 'lapack'],
                                  extra_compile_args = ['-w', '-O3', '-std=c++11', '-stdlib=libc++'])

            self.locals['start_iteration'] += self.locals['numiter']
            self.current_iteration += self.locals['numiter']

            if not self.callback(self.chainnum, self.current_iteration, self.locals):
                break

class eSTAT(object):
     
    def __init__(self, template):
        self.params = []
        self.paramnames = []
        self.customsteps = []
        self.monitor = []
        self.dsets = {}
        self.data = {}
        self.vars = []
        self.results = []

        self.current_iteration = []
        self.template = template
        self.eng = Demo(self.output_handler)

    def output_handler(self, name, ext, f):
        if ext == '.xml':
            if name in self.paramnames and not name in self.customsteps:
                plengths = []
                for p in self.params[0].keys():
                    plengths.append('length(' + p + ')')
                datanames = list(set(self.data.keys()) | set(plengths))
                mode = 0
                needsadapt = False
                out = Parameter(self, name, mode, needsadapt, datanames)
                out.load(f.read())
                return out
            elif name != "model":
                parser = etree.XMLParser(remove_blank_text=True)
                xml = etree.parse(f, parser)
                CtoPtransform = etree.XSLT(etree.parse(os.path.join(xsltpath, 'C to P.xslt')))
                presentation = CtoPtransform(xml)
                return XMLOutput(etree.tostring(presentation), description = 'XML Output')
                #return XMLOutput(minidom.parse(f).toxml(), 'XML Output')

        return None

    def MethodInput(self):
        self.template.EstObjects['nchains']=Integer('Number of chains: ', min=1, default=3)
        self.template.EstObjects['seed']=Text('Random Seed: ', default=1)
        self.template.EstObjects['burnin']=Integer('Length of burnin: ', min=0, default=500)
        self.template.EstObjects['iterations']=Integer('Number of iterations: ', min=0, default=2000, help='The number of iterations post-burnin (not total chain iterations).')
        self.template.EstObjects['thinning']=Integer('Thinning: ', min=1, default=1)
        self.template.EstObjects['defaultalg']=Boolean('Use default algorithm settings: ')
        if self.template.EstObjects['defaultalg']:
            self.template.EstObjects['identifygibbs']=Boolean(value=True)
            self.template.EstObjects['adaptiter']=Integer(value=5000)
            self.template.EstObjects['adaptaccept']=Text(value="44.1")
            self.template.EstObjects['adapttol']=Text(value="10")
        else:
            self.template.EstObjects['identifygibbs']=Boolean('Use Gibbs sampling where possible: ')
            self.template.EstObjects['adaptiter']=Integer('Maximum adapting iterations: ', min=0, default=5000)
            self.template.EstObjects['adaptaccept']=Text('Adapting acceptance rate: ', default="44.1")
            self.template.EstObjects['adapttol']=Text('Adapting tolerance: ',default="10")
        self.template.EstObjects['makepred']=Boolean('Generate prediction dataset: ', default=False)
        self.template.EstObjects['defaultsv']=Boolean('Use default starting values: ')
        if not self.template.EstObjects['defaultsv']:
            for i in range(0, int(self.template.EstObjects['nchains'])):
                for var in self.template.objects:
                    if isinstance(self.template.objects[var], ParamScalar) and self.template.objects[var].modelled == True:
                        self.template.EstObjects[var + '_sv_' + str(i)] = Text('Chain ' + str(i) +': Starting value for ' + var + ' : ', default="0.1")
                    if isinstance(self.template.objects[var], ParamVector) and self.template.objects[var].modelled == True:
                        initval = numpy.ones(numpy.abs(self.template.objects[var].ncols)) * 0.1
                        self.template.EstObjects[var + '_sv_' + str(i)] = List('Chain ' + str(i) +': Starting values for ' + var + '(of length ' + str(len(initval)) + '): ', default=','.join([str(x) for x in initval]))
                    if isinstance(self.template.objects[var], ParamMatrix) and self.template.objects[var].modelled == True:
                        initval = numpy.zeros((self.template.objects[var].size*(self.template.objects[var].size+1))/2)
                        for j in range(0, self.template.objects[var].size):
                            initval[(j * (j + 1) / 2) + j] = 1
                        self.template.EstObjects[var + '_sv_' + str(i)] = List('Chain ' + str(i) +': Starting values for ' + var + '(of length ' + str(len(initval)) + '): ', default=','.join([str(x) for x in initval]))

        canimpute = False
        for var in self.template.objects:
            if var.startswith('mis') and isinstance(self.template.objects[var], ParamVector):
                canimpute = True

        if canimpute == True:
            self.template.EstObjects['imputeiters'] = List('Impute at iterations: ')

        self.template.EstObjects['outdata']=Text('Name of output results: ', help='Outputted data (e.g. chains) will appear under this name in the list of datasets accessible via the menu in the top bar (note these are saved in <em>temporary</em> memory, but can, of course, be downloaded).')

    def create_map(self, a):
        uniqa = numpy.unique(a)
        amap = zip(range(0, len(uniqa)), uniqa)
        d = {}
        for i in amap:
            d[i[1]] = i[0]
        return d

    def applyparams(self, chainnum):
        environ = {}
        for dta in self.dsets:
            for var in self.dsets[dta].variables:
                if dta == 'datafile':
                    environ[var] = self.dsets['datafile'].variables[var]['data']
                else:
                    environ[dta + '_' + var] = self.dsets[dta].variables[var]['data']

        for var in self.template.EstObjects:
            environ[var] = self.template.EstObjects[var]

        environ['chainnum'] = chainnum


        self.paramnames = []
        for var in self.template.objects:
            if isinstance(self.template.objects[var], ParamScalar):
                if self.template.objects[var].customstep == True:
                    self.customsteps.append(var)
                if self.template.objects[var].monitor == True:
                    self.monitor.append(var)
                self.paramnames.append(var)
                initval = 0.1
                if self.template.objects[var].answered == True and self.template.objects[var]._name != "":
                    initval = numpy_eval(str(self.template.objects[var]._name), environ)
                if var + '_sv_' + str(chainnum) in self.template.EstObjects.keys() and self.template.EstObjects[var + '_sv_' + str(chainnum)].answered == True:
                    initval = numpy_eval(str(self.template.EstObjects[var + '_sv_' + str(chainnum)]), environ)
                logging.info('initial values: ' + str(initval))
                self.params[chainnum][var] = initval

            if isinstance(self.template.objects[var], ParamVector):
                if self.template.objects[var].customstep == True:
                    if self.template.objects[var].ncols < 0:
                        self.customsteps.append(var)
                    for i in range(0, self.template.objects[var].ncols):
                        self.customsteps.append(var + '_' + str(i))

                if self.template.objects[var].monitor == True:
                    self.monitor.append(var)
                    for i in range(0, self.template.objects[var].ncols):
                        self.monitor.append(var + str(i))

                if self.template.objects[var].ncols < 0:
                    self.paramnames.append(var)
                else:
                    for i in range(0, self.template.objects[var].ncols):
                        self.paramnames.append(var + '_' + str(i))

                initval = numpy.ones(numpy.abs(self.template.objects[var].ncols)) * 0.1
                if self.template.objects[var].answered == True and self.template.objects[var]._name != "":
                    initval[:] = flatten_list([numpy_eval(str(val), environ) for val in self.template.objects[var]._name])
                if var + '_sv_' + str(chainnum) in self.template.EstObjects and self.template.EstObjects[var + '_sv_' + str(chainnum)].answered == True:
                    initval[:] = flatten_list([numpy_eval(str(val), environ) for val in self.template.EstObjects[var + '_sv_' + str(chainnum)] ])
                if var.startswith('mis'):
                    if var.startswith('miss'):
                        datavar = var[4:]
                    if var.startswith('misb'):
                        datavar = var[4:]
                    if var.startswith('miso'):
                        datavar = var[4:]
                    if var.startswith('misu'):
                        temp = var.rsplit('_', 2)
                        datavar = temp[1] + '_' + temp[2]
                        if temp[0].find('_') != -1:
                            datavar = temp[0].rsplit('_', 1)[0][4:] + '_' + datavar

                    dataindices = (self.data[datavar] > -9.999e29)
                    initval[dataindices] = self.data[datavar][dataindices]

                logging.info('initial values: ' + str(initval))
                self.params[chainnum][var] = initval

            if isinstance(self.template.objects[var], ParamMatrix):
                if self.template.objects[var].customstep == True:
                    self.customsteps.append(var)
                if self.template.objects[var].monitor == True:
                    self.monitor.append(var)
                self.paramnames.append(var)

                initval = numpy.zeros((self.template.objects[var].size*(self.template.objects[var].size+1))/2)
                for i in range(0, self.template.objects[var].size):
                    initval[(i * (i + 1) / 2) + i] = 1
                if self.template.objects[var].answered == True and self.template.objects[var]._name != "":
                    initval[:] = flatten_list([numpy_eval(str(val), environ) for val in self.template.objects[var]._name])
                if var + '_sv_' + str(chainnum) in self.template.EstObjects and self.template.EstObjects[var + '_sv_' + str(chainnum)].answered == True:
                    initval[:] = flatten_list([numpy_eval(str(val), environ) for val in self.template.EstObjects[var + '_sv_' + str(chainnum)] ])
                logging.info('initial values: ' + str(initval))
                self.params[chainnum][var] = initval

    def applydata(self):
        imputevars = []
        for var in self.template.objects:
            if isinstance(self.template.objects[var], ParamVector):
                if var.startswith('mis'):
                    if var.startswith('miss') or var.startswith('misb') or var.startswith('miso'):
                        imputevars.append(var[4:])
                    if var.startswith('misu'):
                        temp = var.rsplit('_', 2)
                        datavar = temp[1]
                        if temp[0].find('_') != -1:
                            datavar = temp[0].rsplit('_', 1)[0][4:] + '_' + datavar
                        imputevars.append(datavar)

        missvars = {k:[] for k in GUIObject.interface.data.keys()}

        needcopy = hasattr(self.template, 'preparedata')
        for var in self.template.objects:
            if isinstance(self.template.objects[var], DataMatrix):
                if self.template.objects[var].categorical != []:
                    needcopy = True
                for datavar in self.template.objects[var].orignames:
                    if self.template.objects[var].data != 'datafile':
                        impvar = self.template.objects[var].data + '_' + datavar
                    else:
                        impvar = datavar
                    if datavar in GUIObject.interface.data[self.template.objects[var].data].variables:
                        if numpy.ma.count_masked(GUIObject.interface.data[self.template.objects[var].data].variables[datavar]['data']) > 0 and not impvar in imputevars:
                            missvars[self.template.objects[var].data].append(datavar)
                            needcopy = True
            if isinstance(self.template.objects[var], DataVector):
                if self.template.objects[var].data != 'datafile':
                    impvar = self.template.objects[var].data + '_' + str(self.template.objects[var])
                else:
                    impvar = str(self.template.objects[var])
                datavar = str(self.template.objects[var])
                if datavar in GUIObject.interface.data[self.template.objects[var].data].variables:
                    if numpy.ma.count_masked(GUIObject.interface.data[self.template.objects[var].data].variables[datavar]['data']) > 0 and not impvar in imputevars:
                        missvars[self.template.objects[var].data].append(datavar)
                        needcopy = True

        self.dsets = {}
        for k in GUIObject.interface.data.keys():
            if needcopy:
                self.dsets[k] = GUIObject.interface.data[k].copy()
            else:
                self.dsets[k] = GUIObject.interface.data[k]

        for var in self.template.objects:
            if isinstance(self.template.objects[var], DataMatrix):
                for catvar in self.template.objects[var].categorical:
                    data = self.dsets[self.template.objects[var].data].variables[catvar]['data']
                    uniqvals = numpy.unique(data.compressed())
                    for j in range(1, len(uniqvals)):
                        self.dsets[self.template.objects[var].data].addvariable(catvar + '_' + str(int(uniqvals[j])), data = numpy.array(data == uniqvals[j], dtype = float))

        if hasattr(self.template, 'preparedata'):
            localvars = self.template.objects.copy()
            localvars['data'] = self.dsets
            localvars['context'] = self.template.objects.copy()
            exec(self.template.preparedata, localvars)

        missmask = {}
        for k in self.dsets:
            missmask[k] = numpy.zeros(self.dsets[k].nobs, dtype=bool)

        for d in missvars:
            for v in missvars[d]:
                missmask[d] = missmask[d] | self.dsets[d].variables[v]['data'].mask
                print('Listwise deleting: ' + v)

        for k in self.dsets:
            self.dsets[k].nobs = sum(missmask[k] == False)
            for var in self.dsets[k].variables.keys():
                self.dsets[k].variables[var]['data'] = self.dsets[k].variables[var]['data'][missmask[k] == False]

        for var in self.template.objects:
            logging.info('var :' + var)
            if isinstance(self.template.objects[var], DataScalar):
                logging.info('var name: ' + self.template.objects[var].name)
                data = self.dsets[self.template.objects[var].data].variables[self.template.objects[var].name]['data']
                if self.template.objects[var].data == 'datafile':
                    self.data[self.template.objects[var].name] = data
                else:
                    self.data[self.template.objects[var].data + '_' + self.template.objects[var].name] = data

            if isinstance(self.template.objects[var], DataVector):
                logging.info('var name: ' + self.template.objects[var].name)
                data = numpy.ma.array(self.dsets[self.template.objects[var].data].variables[self.template.objects[var].name]['data'], dtype = float)
                data = numpy.array(data.filled(fill_value=-9.999e29))
                if self.template.objects[var].data == 'datafile':
                    self.data[self.template.objects[var].name] = data
                else:
                    self.data[self.template.objects[var].data + '_' + self.template.objects[var].name] = data

            if isinstance(self.template.objects[var], IDVector):
                logging.info('ID name: ' + self.template.objects[var].name)
                a = self.dsets[self.template.objects[var].data].variables[self.template.objects[var].name]['data']
                m = self.create_map(a.compressed())
                c = numpy.zeros(len(a), dtype=int)
                for i in range(0, len(a)):
                    if a.mask[i] == False:
                        c[i] = m[a[i]]
                    else:
                        c[i] = -1
                if self.template.objects[var].data == 'datafile':
                    self.data[self.template.objects[var].name] = c
                else:
                    self.data[self.template.objects[var].data + '_' + self.template.objects[var].name] = c

            if isinstance(self.template.objects[var], DataMatrix):
                for i in range(0, len(self.template.objects[var].name)):
                    logging.info('dataset name: ' + self.template.objects[var].data)
                    logging.info('var name: ' + self.template.objects[var].name[i])
                    data = numpy.ma.array(self.dsets[self.template.objects[var].data].variables[self.template.objects[var].name[i]]['data'], dtype = float)
                    data = numpy.array(data.filled(fill_value=-9.999e29))
                    if self.template.objects[var].data == 'datafile':
                        self.data[self.template.objects[var].name[i]] = data
                    else:
                        self.data[self.template.objects[var].data + '_' + self.template.objects[var].name[i]] = data

            if isinstance(self.template.objects[var], IDMatrix):
                a = []
                for i in range(0, self.template.objects[var].ncols()):
                    logging.info('ID name: ' + self.template.objects[var].name[i])
                    a += list(self.dsets[self.template.objects[var].data].variables[self.template.objects[var].name[i]]['data'].compressed())
                m = self.create_map(a)
                for i in range(0, self.template.objects[var].ncols()):
                    b = self.dsets[self.template.objects[var].data].variables[self.template.objects[var].name[i]]['data']
                    c = numpy.zeros(len(b), dtype=int)
                    for j in range(0, len(b)):
                        if b.mask[j] == False:
                            c[j] = m[b[j]]
                        else:
                            c[j] = -1
                    if self.template.objects[var].data == 'datafile':
                        self.data[self.template.objects[var].name[i]] = c
                    else:
                        self.data[self.template.objects[var].data + '_' + self.template.objects[var].name[i]] = c

            #if isinstance(self.template.objects[var], Integer):
            #    self.data[var] = int(self.template.objects[var])
            #if isinstance(self.template.objects[var], Text):
            #    self.data[var] = str(self.template.objects[var])

    def calcconsts(self):
        exprdict = OrderedDict()
        sumdict = {}
        constdata = {}
        sel_count = 0
        extrvectdict = OrderedDict()

        code = self.data_header_code()

        optimisations = []
        try:
            unaryminus = EStat.configuration.getboolean('EStat', 'unaryminus')
        except:
            unaryminus = True
        if unaryminus:
            optimisations.append('unaryminus')

        try:
            joinplus1 = EStat.configuration.getboolean('EStat', 'joinplus1')
        except:
            joinplus1 = True
        if joinplus1:
            optimisations.append('joinplus1')

        try:
            splitdivide = EStat.configuration.getboolean('EStat', 'splitdivide')
        except:
            splitdivide = True
        if splitdivide:
             optimisations.append('splitdivide')

        try:
            expandpowers = EStat.configuration.getboolean('EStat', 'expandpowers')
        except:
            expandpowers = True
        if expandpowers:
            optimisations.append('expandpowers')

        try:
            joinplus2 = EStat.configuration.getboolean('EStat', 'joinplus2')
        except:
            joinplus2 = True
        if joinplus2:
            optimisations.append('joinplus2')

        try:
            jointimes = EStat.configuration.getboolean('EStat', 'jointimes')
        except:
            jointimes = True
        if jointimes:
            optimisations.append('jointimes')

        try:
            expandterms = EStat.configuration.getboolean('EStat', 'expandterms')
        except:
            expandterms = True
        if expandterms:
            optimisations.append('expandterms')

        try:
            splitsums = EStat.configuration.getboolean('EStat', 'splitsums')
        except:
            splitsums = True
        if splitsums:
            optimisations.append('splitsums')

        try:
            splitpartialconst = EStat.configuration.getboolean('EStat', 'splitpartialconst')
        except:
            splitpartialconst = True
        if splitpartialconst:
            optimisations.append('splitpartialconst')

        try:
            mergecommon = EStat.configuration.getboolean('EStat', 'mergecommon')
        except:
            mergecommon = True
        if mergecommon:
            optimisations.append('mergecommon')

        try:
            exprule = EStat.configuration.getboolean('EStat', 'exprule')
        except:
            exprule = True

        if exprule:
            optimisations.append('exprule')

        try:
            removeunaryplus = EStat.configuration.getboolean('EStat', 'removeunaryplus')
        except:
            removeunaryplus = True
        if removeunaryplus:
            optimisations.append('removeunaryplus')

        try:
            mergepowers = EStat.configuration.getboolean('EStat', 'mergepowers')
        except:
            mergepowers = True
        if mergepowers:
            optimisations.append('mergepowers')

        try:
            replaceconst = EStat.configuration.getboolean('EStat', 'replaceconst')
        except:
            replaceconst = True
        if replaceconst:
            optimisations.append('replaceconst')

        for k in self.paramnames:
            logging.debug('Param: ' + k)
            if not k in self.customsteps:
                if k + '.xml' in self.eng.outputs:
                    t = XMLtoSubC(self.eng.outputs[k + '.xml'].tree, exprdict, extrvectdict, sumdict, sel_count, k, optimisations)
                    self.eng.outputs[k + '.xml'].tree = t.extractConsts()
                    sel_count = t.sel_count
                    code += str(t.extracode)

        for e in extrvectdict.keys():
            newlen = 0
            if extrvectdict[e][1] in self.params[0].keys():
                newlen = len(self.params[0][extrvectdict[e][1]])
            elif extrvectdict[e][1] in self.data.keys():
                newlen = len(self.data[extrvectdict[e][1]])
            self.data[extrvectdict[e][0]] = numpy.zeros(newlen)
            code += 'for (int ' + extrvectdict[e][2] + ' = 0; ' + extrvectdict[e][2] + ' < length(' + extrvectdict[e][1] + '); ' + extrvectdict[e][2] + '++) {\n'
            code += '\t' + extrvectdict[e][0] + '[' + extrvectdict[e][2] + '] = ' + extrvectdict[e][3] + ';\n'
            code += '}\n'

        for e in exprdict.keys():
            code += str(exprdict[e][0]) + ' = ' + str(exprdict[e][1]) + ';\n'
            code += "\tpy_%s = PyFloat_FromDouble(%s);\n" % (exprdict[e][0], exprdict[e][0])
            code += "\tPyDict_SetItemString(raw_locals, \"%s\", py_%s);\n" % (exprdict[e][0], exprdict[e][0])
            code += "\tPy_DECREF(py_%s);\n" % (exprdict[e][0])
            constdata[exprdict[e][0]] = 0.0

        for d in self.data.keys():
            if isinstance(self.data[d], list) or isinstance(self.data[d], numpy.ndarray):
                code = code.replace('length(' + d + ')', str(len(self.data[d])))
        for p in self.params[0].keys():
            if isinstance(self.params[0][p], list) or isinstance(self.params[0][p], numpy.ndarray):
                code = code.replace('length(' + p + ')', str(len(self.params[0][p])))

        self.eng.outputs['constcode.cpp'] = TextOutput(code, description = 'Constant optimisation code')

        self.runcode(str(code), constdata, self.data)

        datamap = {}
        for d in self.data.keys():
            if not d.startswith('exprvect'):
                datamap[d] = []

        for d2 in self.data.keys():
            if d2.startswith('exprvect'):
                for d in datamap.keys():
                    if numpy.all(self.data[d] == self.data[d2]):
                        datamap[d].append(d2)
                        break

        vectreplace = []
        for d in datamap.keys():
            for d2 in datamap[d]:
                vectreplace.append((d2, d))
                del self.data[d2]

        for k in self.paramnames:
            if not k in self.customsteps:
                if k + '.xml' in self.eng.outputs:
                    t = XMLtoSubC(self.eng.outputs[k + '.xml'].tree, exprdict, extrvectdict, sumdict, sel_count, k, optimisations)
                    t.replacevar(self.eng.outputs[k + '.xml'].tree, vectreplace)
                    self.eng.outputs[k + '.xml'].tree = t.applyConsts(constdata, self.params[0])
                    if k == 'deviance':
                        desc = "Deviance function"
                    elif self.eng.outputs[k + '.xml'].deterministic:
                        desc = "Deterministic formula for parameter %s" % (k)
                    elif self.eng.outputs[k + '.xml'].needsadapt:
                        desc = "Conditional posterior for %s for random walk Metropolis" % (k)
                    else:
                        desc = "Conditional posterior for %s for Gibbs sampling" % (k)
                    self.eng.outputs[k + '.xml'].consttex = MathOutput(self.eng.outputs[k + '.xml'].asLaTeX(), description = desc)

    def compilemodel(self):
        includes = '<%! from EStat.Utils import mmult, mmult2, index, create_map, make_class, bracket, combs, combsi, combsi2 %>'
        modelspec = MakoTemplate(includes + self.template.model).render(**self.template.objects)
        logging.info(modelspec)

        if modelspec != '':
            self.eng.inputs['model.txt'] = TextOutput(modelspec, description = 'Model specification code')

            mylex = lex.lex(module=bitbmsl)
            myyacc = yacc.yacc(module=bitbmsl, debug=0, write_tables=0)
            #mylex = lex.lex(module=bitbmsl, optimize=1)
            #myyacc = yacc.yacc(module=bitbmsl, optimize=1, debug=0)
            parser = etree.XMLParser(remove_blank_text=True)
            modelxml = None
            try:
                modelxml = etree.XML(myyacc.parse(modelspec,lexer=mylex), parser)
            except ValueError:
                print('Error parsing:\n' + modelspec)

            if modelxml != None:
                self.eng.inputs['model.xml'] = TextOutput(etree.tostring(modelxml), description = 'Model specification tree')

                CtoPtransform = etree.XSLT(etree.parse(os.path.join(xsltpath, 'C to P.xslt')))
                presentation = CtoPtransform(modelxml)
                self.eng.outputs['presentation.xml'] = XMLOutput(etree.tostring(presentation), description = 'Presentation MathML rendering of model')

                #PtoCtransform = etree.XSLT(etree.parse(os.path.join(xsltpath, 'P to C.xslt')))
                #content = PtoCtransform(presentation)
                #self.eng.outputs['content.xml'] = XMLOutput(etree.tostring(content), description = 'Content MathML rendering of model')

                CtoVGMtransform = etree.XSLT(etree.parse(os.path.join(xsltpath, 'C to VGM P.xslt')))
                vgm = CtoVGMtransform(modelxml)
                print(CtoVGMtransform.error_log)
                self.eng.outputs['vgm.xml'] = XMLOutput(etree.tostring(vgm), description = 'VGM rendering of model')

                CtoBUGStransform = etree.XSLT(etree.parse(os.path.join(xsltpath, 'C to BUGS.xslt')))
                bugs = CtoBUGStransform(modelxml)
                self.eng.outputs['bugs.xml'] = XMLOutput(etree.tostring(bugs), description = 'BUGS rendering of model')

                self.eng.identify = bool(self.template.EstObjects['identifygibbs'])
                self.eng.run('model.xml')

    def find_adapt(self, chainnum):
        needadapt = {}
        for var in self.template.objects:
            if isinstance(self.template.objects[var], ParamScalar):
                if var in self.customsteps or (var + '.xml' in self.eng.outputs and self.eng.outputs[var + '.xml'].needsadapt == True):
                    needadapt[var] = self.params[chainnum][var]
            if isinstance(self.template.objects[var], ParamVector):
                adaptvar = False
                if self.template.objects[var].ncols < 0:
                    if var in self.customsteps or (var + '.xml' in self.eng.outputs and self.eng.outputs[var + '.xml'].needsadapt == True):
                        adaptvar = True
                else:
                    for i in range(0, self.template.objects[var].ncols):
                        if var + '_' + str(i) in self.customsteps or (var + '_' + str(i) + '.xml' in self.eng.outputs and self.eng.outputs[var + '_' + str(i) + '.xml'].needsadapt == True):
                            adaptvar = True
                if adaptvar == True:
                    needadapt[var] = self.params[chainnum][var]
            if isinstance(self.template.objects[var], ParamMatrix):
                 if var in self.customsteps or (var + '.xml' in self.eng.outputs and self.eng.outputs[var + '.xml'].needsadapt == True):
                     needadapt[var] = self.params[chainnum][var]
        return needadapt

    def data_header_code(self):
        try:
            req_classifications=[]
            for k in self.template.objects:
                if isinstance(self.template.objects[k], IDVector):
                    if self.template.objects[k].data == 'datafile':
                        req_classifications.append(self.template.objects[k].name)
                    else:
                        req_classifications.append(self.template.objects[k].data + '_' + self.template.objects[k].name)
                if isinstance(self.template.objects[k], IDMatrix):
                    if self.template.objects[k].data == 'datafile':
                        req_classifications.append(self.template.objects[k].name)
                    else:
                        req_classifications.append(self.template.objects[k].data + '_' + self.template.objects[k].name)

            for c in req_classifications:
                if isinstance(c, list):
                    d = set()
                    for x in c:
                        d.update(self.data[x])
                    for x in c:
                        if 'idmap_' + x not in self.data:
                            mx = len(numpy.unique(d)) #len(self.params[0][k])
                            idmap=[]
                            for i in range(mx):
                                idmap.append([])
                            numids = 0
                            for i in range(len(self.data[x])):
                                if self.data[x][i] != -1:
                                    idmap[int(self.data[x][i])].append(i)
                                    numids += 1

                            idmap_sizes = numpy.zeros(len(idmap), dtype = int)
                            idmap_data = numpy.zeros(numids, dtype = int)
                            n = 0
                            for i in range(0, len(idmap)):
                                idmap_sizes[i] = len(idmap[i])
                                idmap_data[n:n+len(idmap[i])] = idmap[i][0:len(idmap[i])]
                                n += len(idmap[i])

                            self.data['idmapsizes_'+ x] = idmap_sizes
                            self.data['idmapdata_'+ x] = idmap_data

                else:
                    d = self.data[c]
                    if 'idmap_' + c not in self.data:
                        mx = len(numpy.unique(d)) #len(self.params[0][k])
                        idmap=[]
                        for i in range(mx):
                            idmap.append([])
                        numids = 0
                        for i in range(len(self.data[c])):
                            if self.data[c][i] != -1:
                                idmap[int(self.data[c][i])].append(i)
                                numids += 1

                        idmap_sizes = numpy.zeros(len(idmap), dtype = int)
                        idmap_data = numpy.zeros(numids, dtype = int)
                        n = 0
                        for i in range(0, len(idmap)):
                            idmap_sizes[i] = len(idmap[i])
                            idmap_data[n:n+len(idmap[i])] = idmap[i][0:len(idmap[i])]
                            n += len(idmap[i])

                        self.data['idmapsizes_'+ c] = idmap_sizes
                        self.data['idmapdata_'+ c] = idmap_data

        except AttributeError:
            pass

        code = ''
        for var in self.template.objects:
            if isinstance(self.template.objects[var], DataMatrix):
                code += 'std::vector<double*> tmp_' + str(var) + ';\n'
                for i in range(0, self.template.objects[var].ncols()):
                    if self.template.objects[var].data == 'datafile':
                        code += 'tmp_' + str(var) + '.push_back(const_cast<double *>(' + str(self.template.objects[var][i]) + '));\n'
                    else:
                        code += 'tmp_' + str(var) + '.push_back(const_cast<double *>(' + self.template.objects[var].data + '_' + str(self.template.objects[var][i]) + '));\n'
                if self.template.objects[var].data == 'datafile':
                    code += 'RectMatrix mat_' + str(var) + '(tmp_' + str(var) + ', length(' + str(self.template.objects[var][0]) + '));\n'
                else:
                    code += 'RectMatrix mat_' + str(var) + '(tmp_' + str(var) + ', length(' + self.template.objects[var].data + '_' + str(self.template.objects[var][0]) + '));\n'
            if isinstance(self.template.objects[var], DataVector):
                code += 'std::vector<double*> tmp_' + str(var) + ';\n'
                if self.template.objects[var].data == 'datafile':
                    code += 'tmp_' + str(var) + '.push_back(const_cast<double *>(' + str(self.template.objects[var]) + '));\n'
                    code += 'RectMatrix mat_' + str(var) + '(tmp_' + str(var) + ', length(' + str(self.template.objects[var]) + '));\n'
                else:
                    code += 'tmp_' + str(var) + '.push_back(const_cast<double *>(' + self.template.objects[var].data + '_' + str(self.template.objects[var]) + '));\n'
                    code += 'RectMatrix mat_' + str(var) + '(tmp_' + str(var) + ', length(' + self.template.objects[var].data + '_' + str(self.template.objects[var]) + '));\n'

        for var in req_classifications:
            if isinstance(var, list):
                for c in var:
                    code += 'std::vector<arrayref> idmap_' + c + '(' + str(len(self.data['idmapsizes_'+ c])) + ');\n'
                    code += 'for (int i = 0, start = 0; i < idmap_' + c + '.size(); i++) {\n'
                    code += '    idmap_' + c + '[i] = arrayref(const_cast<long int *>(&idmapdata_' + c + '[start]), idmapsizes_' + c + '[i]);\n'
                    code += '    start += idmapsizes_' + c + '[i];\n'
                    code += '}\n'
            else:
                code += 'std::vector<arrayref> idmap_' + var + '(' + str(len(self.data['idmapsizes_'+ var])) + ');\n'
                code += 'for (int i = 0, start = 0; i < idmap_' + var + '.size(); i++) {\n'
                code += '    idmap_' + var + '[i] = arrayref(const_cast<long int *>(&idmapdata_' + var + '[start]), idmapsizes_' + var + '[i]);\n'
                code += '    start += idmapsizes_' + var + '[i];\n'
                code += '}\n'

        return code

    def param_header_code(self):
        code = ''
        for var in self.template.objects:
            if isinstance(self.template.objects[var], ParamMatrix):
                code += 'SymMatrix mat_' + str(var) + '(' + str(var) + ', ' + str(self.template.objects[var].size) + ');\n'
            if isinstance(self.template.objects[var], ParamVector):
                code += 'std::vector<double*> tmp_' + str(var) + ';\n'
                code += 'tmp_' + str(var) + '.push_back(' + str(var) + ');\n'
                code += 'RectMatrix mat_' + str(var) + '(tmp_' + str(var) + ', length(' + str(var) + '));\n'
                if self.template.objects[var].ncols > 0:
                    for i in range(0, self.template.objects[var].ncols):
                        code += 'double &' + str(var) + '_' + str(i) + ' = ' + str(var) + '[' + str(i) + '];\n'
        return code

    def adapt_header_code(self):
        code = ''
        for var in self.template.objects:
            if isinstance(self.template.objects[var], ParamVector):
                if self.template.objects[var].ncols > 0:
                    if var in self.needadapt.keys():
                        for i in range(0, self.template.objects[var].ncols):
                            code += 'double &' + str(var) + '_' + str(i) + '_sd' + ' = ' + str(var) + '_sd' + '[' + str(i) + '];\n'
                        for i in range(0, self.template.objects[var].ncols):
                            code += 'long int &' + str(var) + '_' + str(i) + '_accept' + ' = ' + str(var) + '_accept' + '[' + str(i) + '];\n'
        return code

    def adapt_header_code2(self):
        code = ''
        for var in self.template.objects:
            if isinstance(self.template.objects[var], ParamVector):
                if self.template.objects[var].ncols > 0:
                    if var in self.needadapt.keys():
                        for i in range(0, self.template.objects[var].ncols):
                            code += 'double &' + str(var) + '_' + str(i) + '_sd' + ' = ' + str(var) + '_sd' + '[' + str(i) + '];\n'
                        for i in range(0, self.template.objects[var].ncols):
                            code += 'int &' + str(var) + '_' + str(i) + '_accept' + ' = ' + str(var) + '_accept' + '[' + str(i) + '];\n'
        return code

    def runcode(self, code, locals, globals):
        if not hasattr(self, 'compiled'):
            self.compiled = []

        if not code in self.compiled:
            lock.acquire()
            if code in self.compiled:
                lock.release()

        try:
            result = scipy.weave.inline(code, 
                                  locals.keys() + globals.keys(),
                                  local_dict = locals, 
                                  global_dict = globals, 
                                  support_code = self.scode, 
                                  compiler = 'gcc', 
                                  verbose = 2,
                                  headers = ['<unordered_map>', '<vector>', '<cmath>', '<cfloat>', '<limits>', '<random>', '"rng.h"', '"statlib.h"'],
                                  include_dirs = [tmplpath],
                                  sources = [os.path.join(tmplpath, 'statlib.cpp'), os.path.join(tmplpath, 'rng_simple.cpp')],
                                  library_dirs = [tmplpath],
                                  libraries = ['blas', 'lapack'],
                                  extra_compile_args = ['-w', '-O3', '-std=c++11' ])

        except:
            logging.debug('Unexpected error:' + str(sys.exc_info()[0]))
            # The code was not successfully compiled, so don't cache it.
            code = ''

        finally:
            if not code in self.compiled:
                if code != '':
                    self.compiled.append(code)
                lock.release()

        if code == '':
            raise

        return result

    def init(self):
        self.params = [{} for i in range(int(self.template.EstObjects['nchains']))]
        self.results = [{} for i in range(int(self.template.EstObjects['nchains']))]
        self.vars = [{} for i in range(int(self.template.EstObjects['nchains']))]
        self.current_iteration = [0 for i in range(int(self.template.EstObjects['nchains']))]

        self.applydata()
        for i in range(int(self.template.EstObjects['nchains'])):
            self.applyparams(i)

        if hasattr(self.template, 'latex'):
            self.eng.outputs['equation.tex'] = self.template.outputlatex()
        else:
            self.eng.outputs['equation.tex'] = MathOutput('\\text{No equation specified for this template}', 'LaTeX equation')

        if hasattr(self.template, 'model'):
            self.compilemodel()

        LaTeXalgorithm = '<script type="text/javascript" src="/static/MathJax/MathJax.js?config=default"></script>'

        if hasattr(self.template, 'customTeX'):
            customTeX = self.template.customTeX()
        else:
            customTeX = {}

        allmissing = True
        for p in self.paramnames:
            if not p in self.customsteps:
                LaTeXalgorithm += '<div style="text-align: left; margin-top: 25px; margin-left: 25px">'
                if p == 'deviance':
                    LaTeXalgorithm += '<p>Deviance Function</p>'
                elif self.eng.outputs[p + '.xml'].deterministic:
                    LaTeXalgorithm += '<p>Deterministic formula for parameter ' + p + '</p>'
                elif self.eng.outputs[p + '.xml'].needsadapt:
                    LaTeXalgorithm += '<p>Conditional posterior for ' + p + ' for random walk Metropolis</p>'
                else:
                    LaTeXalgorithm += '<p>Conditional posterior for ' + p + ' for Gibbs sampling</p>'
                LaTeXalgorithm += '\\['
                if p + '.xml' in self.eng.outputs:
                    LaTeXalgorithm += self.eng.outputs[p + '.xml'].asLaTeX()
                else:
                    LaTeXalgorithm += 'Not available'
                LaTeXalgorithm += '\\]</div>'
                allmissing = False
            else:
                if p in customTeX:
                    LaTeXalgorithm += '<div style="text-align: left; margin-top: 25px; margin-left: 25px">'
                    from mako import exceptions
                    try:
                        tex = MakoTemplate(customTeX[p]).render(**self.template.objects)
                    except:
                        tex = exceptions.text_error_template().render()
                    LaTeXalgorithm += '<p>Custom step for ' + p + '</p>'
                    LaTeXalgorithm += '\\['
                    LaTeXalgorithm += tex
                    LaTeXalgorithm += '\\]</div>'
                else:
                    LaTeXalgorithm += '<div style="text-align: left; margin-top: 25px; margin-left: 10px">Parameter ' + p +' is updated using custom user defined step</div>'
                    logging.info('Custom step needed for parameter: ' + p)
        if allmissing == True and self.customsteps == []:
            logging.error("Error in algebra system, nothing generated")
            return None


        self.eng.outputs['algorithm.tex'] = HTMLOutput(LaTeXalgorithm, description = 'LaTeX version of algorithm')

        self.scode = '''
int length(PyObject *obj){
    return Py_SIZE(obj);
}
'''
        self.eng.outputs['supportcode.cpp'] = TextOutput(self.scode, description = 'Statistical support code')

        try:
            optimise = EStat.configuration.getboolean('EStat', 'optimisecode')
        except:
            optimise = True

        if optimise == True:
            self.calcconsts()

        self.needadapt = self.find_adapt(0)

        desacc = float(self.template.EstObjects['adaptaccept']) # Desire x% acceptance
        tol = float(self.template.EstObjects['adapttol']) # with tolerance of x%

        adaptvars = {}
        adaptvars['needadapt'] = self.needadapt
        adaptvars['desacc'] = desacc
        adaptvars['tol'] = tol
        adaptcode = stdtmpl.get_template('adapt.cpp').render(**adaptvars)

        self.eng.outputs['adapt.cpp'] = TextOutput(adaptcode, description = 'Model adapting code')

        deviancecode = ''
        if not 'deviance' in self.customsteps and 'deviance.xml' in self.eng.outputs.keys():
            deviancecode += self.eng.outputs['deviance.xml'].asC(stdtmpl)
        else:
            if hasattr(self.template, 'deviancecode') or hasattr(self.template, 'diccode'):
                if hasattr(self.template, 'diccode'):
                    deviancecode += self.template.diccode
                else:
                    deviancecode += self.template.deviancecode

        if deviancecode != '':
            diccode = self.data_header_code()
            diccode += self.param_header_code()

            diccode += deviancecode
            diccode += "\tpy_deviance = PyFloat_FromDouble(deviance);\n"
            diccode += "\tPyDict_SetItemString(raw_locals, \"deviance\", py_deviance);\n"
            diccode += "\tPy_DECREF(py_deviance);\n"
            for d in self.data.keys():
                if isinstance(self.data[d], list) or isinstance(self.data[d], numpy.ndarray):
                    diccode = diccode.replace('length(' + d + ')', str(len(self.data[d])))
            for p in self.params[0].keys():
                if isinstance(self.params[0][p], list) or isinstance(self.params[0][p], numpy.ndarray):
                    diccode = diccode.replace('length(' + p + ')', str(len(self.params[0][p])))

            from mako import exceptions
            try:
                diccode = MakoTemplate(diccode).render(**self.template.objects)
            except:
                diccode = exceptions.text_error_template().render()
                diccode += str(self.template.objects)

            for d in self.data.keys():
                if isinstance(self.data[d], list) or isinstance(self.data[d], numpy.ndarray):
                    diccode = diccode.replace('length(' + d + ')', str(len(self.data[d])))
            for p in self.params[0].keys():
                if isinstance(self.params[0][p], list) or isinstance(self.params[0][p], numpy.ndarray):
                    diccode = diccode.replace('length(' + p + ')', str(len(self.params[0][p])))

            self.eng.outputs['dic.cpp'] = TextOutput(diccode, description = 'DIC calculation code')

        code = ''
        for k in self.params[0].keys():
            if not k in self.monitor:
                if not isinstance(self.params[0][k], numpy.ndarray):
                    code += 'RunningStat %s_results(%s_n, %s_oldM, %s_newM, %s_oldS, %s_newS);\n' % (k, k, k, k, k, k)
                else:
                    code += 'RunningStatVector %s_results(%d, %s_n, %s_oldM, %s_newM, %s_oldS, %s_newS);\n' % (k, len(self.params[0][k]), k, k, k, k, k)

        code += 'Py_BEGIN_ALLOW_THREADS;\n'
        code += self.data_header_code()
        code += self.param_header_code()
        code += self.adapt_header_code()

        seed = int(self.template.EstObjects['seed'])

        code += 'static std::unordered_map<std::string, rng_t> rngstate;\n'
        code += 'rng_t rng;\n'
        code += 'if (runstate ==0) {\n'
        code += '\trng=rng_create(1,1);\n'
        code += '\trng_set_seed(rng, seed);\n'
        code += '\trng_start(rng);\n'
        code += '\trngstate[rngid]=rng;\n'
        code += '} else {\n'
        code += '\trng=rngstate[rngid];\n'
        code += '}\n'

        code += 'for (int iter = 0; iter < numiter; iter++) {\n'
        code += '\tint iterind = 0;\n'
        code += '\tif (runstate == 3) iterind = start_iteration + floor(iter / thinning);\n'

        if hasattr(self.template, 'preccode'):
            code += self.template.preccode

        for k in self.paramnames:
            if k != 'deviance' and not k in self.customsteps:
                code += '\t// Update ' + k + '\n'
                code += self.eng.outputs[k + '.xml'].asC(stdtmpl)
                code += '\n'

        code += '\t// Update deviance\n'
        if not 'deviance' in self.customsteps and 'deviance.xml' in self.eng.outputs.keys():
            code += self.eng.outputs['deviance.xml'].asC(stdtmpl)

        if hasattr(self.template, 'deviancecode'):
            code += self.template.deviancecode
            code += '\n'

        if hasattr(self.template, 'postccode'):
            code += self.template.postccode

        code += "\tif (runstate == 3) {"
        code += "\t//Store current value into chain\n"
        for k in self.params[0].keys():
            if k in self.monitor:
                if not isinstance(self.params[0][k], numpy.ndarray):
                    code += "\t\t%s_chain[iterind] = %s;\n" % (k,k)
                else:
                    code += "\t\tfor (int i = 0; i < %d; ++i) {\n" % (len(self.params[0][k]))
                    code += "\t\t\t%s_chain[(i * (S%s_chain[0] / sizeof(double))) + iterind * (S%s_chain[1] / sizeof(double))] = %s[i];\n" % (k, k, k, k)
                    code += "\t\t}\n"
            else:
                if not isinstance(self.params[0][k], numpy.ndarray):
                    code += "\t\t%s_results.Push(%s);\n" % (k, k)
                else:
                    code += "\t\t%s_results.Push(%s);\n" % (k, k)

        code += '\t}\n'
        code += '}\n'
        code += 'Py_END_ALLOW_THREADS;\n'

        for k in self.params[0].keys():
            if not k in self.monitor:
                if not isinstance(self.params[0][k], numpy.ndarray):
                    code += '\t%s_results.GetState(%s_n, %s_oldM, %s_newM, %s_oldS, %s_newS);\n' % (k, k, k, k, k, k)
                    code += "\tpy_%s_n = PyInt_FromLong(%s_n);\n" % (k, k)
                    code += "\tPyDict_SetItemString(raw_locals, \"%s_n\", py_%s_n);\n" % (k, k)
                    code += "\tPy_DECREF(py_%s_n);\n" % (k)
                    code += "\tpy_%s_oldM = PyFloat_FromDouble(%s_oldM);\n" % (k, k)
                    code += "\tPyDict_SetItemString(raw_locals, \"%s_oldM\", py_%s_oldM);\n" % (k, k)
                    code += "\tPy_DECREF(py_%s_oldM);\n" % (k)
                    code += "\tpy_%s_newM = PyFloat_FromDouble(%s_newM);\n" % (k, k)
                    code += "\tPyDict_SetItemString(raw_locals, \"%s_newM\", py_%s_newM);\n" % (k, k)
                    code += "\tPy_DECREF(py_%s_newM);\n" % (k)
                    code += "\tpy_%s_oldS = PyFloat_FromDouble(%s_oldS);\n" % (k, k)
                    code += "\tPyDict_SetItemString(raw_locals, \"%s_oldS\", py_%s_oldS);\n" % (k, k)
                    code += "\tPy_DECREF(py_%s_oldS);\n" % (k)
                    code += "\tpy_%s_newS = PyFloat_FromDouble(%s_newS);\n" % (k, k)
                    code += "\tPyDict_SetItemString(raw_locals, \"%s_newS\", py_%s_newS);\n" % (k, k)
                    code += "\tPy_DECREF(py_%s_newS);\n" % (k)
                else:
                    code += '\t%s_results.GetState(%s_n, %s_oldM, %s_newM, %s_oldS, %s_newS);\n' % (k, k, k, k, k, k)
                    code += "\tpy_%s_n = PyInt_FromLong(%s_n);\n" % (k, k)
                    code += "\tPyDict_SetItemString(raw_locals, \"%s_n\", py_%s_n);\n" % (k, k)

        code += "\t//Floats and Ints in python are immutable, so create new ones for the updated values\n"
        for k in self.params[0].keys():
            if not isinstance(self.params[0][k], numpy.ndarray):
                code += "\tpy_%s = PyFloat_FromDouble(%s);\n" % (k, k)
                code += "\tPyDict_SetItemString(raw_locals, \"%s\", py_%s);\n" % (k, k)
                code += "\tPy_DECREF(py_%s);\n" % (k)

        for k in self.needadapt.keys():
            if not isinstance(self.needadapt[k], numpy.ndarray):
                code += "\tpy_%s_accept = PyInt_FromLong(%s_accept);\n"% (k, k)
                code += "\tPyDict_SetItemString(raw_locals, \"%s_accept\", py_%s_accept);\n"% (k, k)
                code += "\tPy_DECREF(py_%s_accept);\n"% (k)

        from mako import exceptions
        try:
            code = MakoTemplate(code).render(**self.template.objects)
        except:
            code = exceptions.text_error_template().render()
            code += str(self.template.objects)

        for d in self.data.keys():
            if isinstance(self.data[d], numpy.ndarray):
                code = code.replace('length(' + d + ')', str(len(self.data[d])))
        for p in self.params[0].keys():
            if isinstance(self.params[0][p], numpy.ndarray):
                code = code.replace('length(' + p + ')', str(len(self.params[0][p])))

        self.eng.outputs['modelcode.cpp'] = TextOutput(code, description = 'Model iteration code')

        try:
            standalone = EStat.configuration.getboolean('EStat', 'use_standalone')
        except:
            standalone = False

        if standalone == True:
            self.eng.outputs['engine.cpp'] = TextOutput(self.genCPP(0), description = 'Standalone engine code') 

    def run(self):
        try:
            standalone = EStat.configuration.getboolean('EStat', 'use_standalone')
        except:
            standalone = False

        if standalone == True:
            self.runCPP()
            return


        def callback(chainnum, iteration, paramsvals):
            if iteration % 100 == 0:
                logging.info('Adapting (chain ' + str(chainnum) + ') for %s iterations (maximum %s)...' % (iteration, int(self.template.EstObjects['adaptiter'])))

                self.runcode(str(self.eng.outputs['adapt.cpp'].content), self.adaptvars[chainnum], self.vars[chainnum])

                num_done = 0
                total_vars = 0
                for k in self.needadapt.keys():
                    if isinstance(self.needadapt[k], numpy.ndarray):
                        for i in range(0, len(self.needadapt[k])):
                            total_vars += 1
                            num_done += self.adaptvars[chainnum][k + '_acceptdone'][i]
                    else:                    
                        total_vars += 1
                        num_done += self.adaptvars[chainnum][k + '_acceptdone']

                if num_done == total_vars:
                    return False

                if iteration >= int(self.template.EstObjects['adaptiter']):
                    return False

            return True

        run_time = time.clock()

        self.adaptvars = [{} for i in range(int(self.template.EstObjects['nchains']))]
 
        adapt = any(self.needadapt) or any(self.customsteps)
        for chainnum in range(0, int(self.template.EstObjects['nchains'])):
            self.vars[chainnum]['chainnum'] = chainnum
            self.vars[chainnum]['numchains'] = int(self.template.EstObjects['nchains'])
            self.vars[chainnum]['seed'] = (chainnum + 1)*int(self.template.EstObjects['seed'])
            for k in self.params[chainnum].keys():
                self.vars[chainnum][k] = self.params[chainnum][k]
                if k in self.monitor:
                    if isinstance(self.vars[chainnum][k], numpy.ndarray):
                        self.results[chainnum][k] = numpy.zeros([len(self.vars[chainnum][k]), 0], order='F')
                    else:
                        self.results[chainnum][k] = numpy.zeros(0)
                    self.vars[chainnum][k+'_chain'] = self.results[chainnum][k]
                else:
                    if isinstance(self.vars[chainnum][k], numpy.ndarray):
                        self.results[chainnum][k] = RunningStatVector(len(self.vars[chainnum][k]))
                    else:
                        self.results[chainnum][k] = RunningStat()
                    self.vars[chainnum][str(k)+'_n'] = self.results[chainnum][k].m_n
                    self.vars[chainnum][str(k)+'_oldM'] = self.results[chainnum][k].m_newM
                    self.vars[chainnum][str(k)+'_newM'] = self.results[chainnum][k].m_oldM
                    self.vars[chainnum][str(k)+'_oldS'] = self.results[chainnum][k].m_newS
                    self.vars[chainnum][str(k)+'_newS'] = self.results[chainnum][k].m_oldS

            for k in self.needadapt.keys():
                if isinstance(self.needadapt[k], numpy.ndarray):
                    self.vars[chainnum][k + '_accept'] = numpy.array([0 for x in self.params[chainnum][k]])
                    self.adaptvars[chainnum][k + '_acceptinrow'] = numpy.array([0 for x in self.params[chainnum][k]])
                    self.adaptvars[chainnum][k + '_acceptdone'] = numpy.array([0 for x in self.params[chainnum][k]])
                    self.vars[chainnum][k + '_sd'] = numpy.array([0.1 for x in self.params[chainnum][k]])
                else:
                    self.vars[chainnum][k + '_accept'] = 0
                    self.adaptvars[chainnum][k + '_acceptinrow'] = 0
                    self.adaptvars[chainnum][k + '_acceptdone'] = 0
                    self.vars[chainnum][k + '_sd'] = 0.1

            logging.info('Finished adapting setup:' + str(time.clock() - run_time))
            run_time = time.clock()

            self.vars[chainnum]['rngid'] = str(uuid.uuid1())
            self.vars[chainnum]['runstate'] = 0
            self.vars[chainnum]['start_iteration'] = 0
            self.vars[chainnum]['numiter'] = 1
            self.vars[chainnum]['thinning'] = 1

            # Run setup code
            self.runcode(str(self.eng.outputs['modelcode.cpp'].content), self.vars[chainnum], self.data)
            if adapt:
                self.runcode(str(self.eng.outputs['adapt.cpp'].content), self.adaptvars[chainnum], self.vars[chainnum])

            for k in self.params[chainnum].keys():
                if isinstance(self.params[chainnum][k], numpy.ndarray):
                    if not numpy.all(numpy.isfinite(self.params[chainnum][k])):
                        logging.error(str(k) + ' contains non finite value, aborting')
                        return
                else:
                    if not numpy.isfinite(self.params[chainnum][k]):
                        logging.error(str(k) + ' has a non finite value, aborting')
                        return

            logging.info('Finished initialising chain ' + str(chainnum) + ' :' + str(time.clock() - run_time))
            run_time = time.clock()

            self.vars[chainnum]['runstate'] = 1
            self.vars[chainnum]['start_iteration'] = 0
            self.vars[chainnum]['thinning'] = 1

        if adapt:
            iterthreads = []
            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                iterthreads.append(ModelThread(callback, chainnum, 1, int(self.template.EstObjects['adaptiter']), range(0, int(self.template.EstObjects['adaptiter']), 100), str(self.eng.outputs['modelcode.cpp'].content), self.scode, self.vars[chainnum], self.data))
                iterthreads[chainnum].start()

            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                iterthreads[chainnum].join()
                self.vars[chainnum]['runstate'] = 3
                logging.info('Finished adapting for chain ' + str(chainnum) + ' :' + str(time.clock() - run_time))

        self.runmore()

    def runmore(self):
        def callback(chainnum, iteration, paramsvals):
            self.current_iteration[chainnum] = iteration

            if iteration in imputeiters:
                imputevars = {}
                for var in self.template.objects:
                    if var.startswith('mis') and isinstance(self.template.objects[var], ParamVector):
                        if var.startswith('miss'):
                            type = 'normal'
                            datavar = var[4:]
                        if var.startswith('misb'):
                            type = 'binary'
                            datavar = var[4:]
                        if var.startswith('miso'):
                            type = 'ordered'
                            datavar = var[4:]
                        if var.startswith('misu'):
                            type = 'unordered'
                            temp = var.rsplit('_', 2)
                            datavar = temp[1]
                            if temp[0].find('_') != -1:
                                datavar = temp[0].rsplit('_', 1)[0][4:] + '_' + datavar

                        dname = 'datafile'
                        for dset in self.dsets.keys():
                            if isinstance(self.dsets[dset], DTAFile):
                                if datavar.startswith(dset):
                                    datavar = datavar.replace(dset + '_', '')
                                    dname = dset
                                    break

                        if not dname in imputevars.keys():
                            imputevars[dname] = []
                        imputevars[dname].append((type, datavar, var))

                    for dset in imputevars.keys():
                        origdata = self.dsets[dset]
                        dta = DTAFile()
                        dta.nobs = origdata.nobs
                        # First copy the original variables
                        for var in origdata.variables.keys():
                            dta.addvariable(var, data = numpy.array(origdata.variables[var]['data']))
                            dta.variables[var]['data'].mask = origdata.variables[var]['data'].mask
                        # Now add the imputed variables
                        logging.info('Imputed missing variables: ' + str(imputevars[dset]))
                        for var in imputevars[dset]:
                            type = var[0]
                            varname = var[1]
                            parname = var[2]
                            missmask = origdata.variables[varname]['data'].mask
                            dta.variables[varname]['data'].mask *= 0

                            if type == 'normal':
                                #dta.addvariable(varname, data = numpy.array(paramsvals[parname]))
                                dta.variables[varname]['data'][missmask==True] = numpy.array(paramsvals[parname])[missmask==True]

                            if type == 'binary':
                                plen = len(paramsvals[parname])
                                temp = numpy.zeros(plen)
                                temp[paramsvals[parname] > 0] = 1
                                #dta.addvariable(varname, data = temp)
                                dta.variables[varname]['data'][missmask==True] = numpy.array(temp)[missmask==True]

                            if type == 'ordered':
                                plen = len(paramsvals[parname])
                                temp = numpy.zeros(plen)
                                cat = 0
                                for j in range(0, len(paramsvals['alpha' + varname])):
                                   cat += 1
                                   temp[paramsvals[parname] > paramsvals['alpha' + varname][j]] = cat
                                #dta.addvariable(varname, data = temp)
                                dta.variables[varname]['data'][missmask==True] = numpy.array(temp)[missmask==True]

                            if type == 'unordered' and parname.endswith('_0'):
                                plen = len(paramsvals[parname])
                                numcats = int(self.template.objects['ncats' + varname])
                                dname = ''
                                if dset != 'datafile':
                                    dname = dset + '_'

                                temp = numpy.ones(plen) * (numcats - 1)
                                for j in range(0, plen):
                                    maxlv = 0.0
                                    for i in range(0, numcats - 1):
                                        if paramsvals['misu' + dname + str(numcats) + '_' + varname + '_' + str(i)][j] > maxlv:
                                            maxlv = paramsvals['misu' + dname + str(numcats) + '_' + varname + '_' + str(i)][j]
                                            temp[j] = i
                                #dta.addvariable(varname, data = temp)
                                dta.variables[varname]['data'][missmask==True] = numpy.array(temp)[missmask==True]

                        self.eng.outputs['impute_' + dset + '_chain' + str(chainnum) + '_iter' + str(iteration-burnin)] = dta

            return True

        burnin = int(self.template.EstObjects['burnin'])
        iterations = int(self.template.EstObjects['iterations'])
        thinning = int(self.template.EstObjects['thinning'])

        imputeiters = []
        if 'imputeiters' in self.template.EstObjects and self.template.EstObjects['imputeiters'].answered == True:
            imputeiters = [burnin + int(i) for i in self.template.EstObjects['imputeiters']]

        total = burnin + iterations

        run_time = time.clock()
        for chainnum in range(0, int(self.template.EstObjects['nchains'])):
            self.vars[chainnum]['chainnum'] = chainnum
            self.vars[chainnum]['seed'] = (chainnum + 1) *int(self.template.EstObjects['seed'])
            self.vars[chainnum]['runstate'] = 2
            self.vars[chainnum]['start_iteration'] = 0
            self.vars[chainnum]['thinning'] = thinning
            for k in self.params[chainnum].keys():
                #self.vars[chainnum][k] = self.params[chainnum][k]
                if k in self.monitor:
                    self.vars[chainnum][k+'_chain'] = None
                    if isinstance(self.vars[chainnum][k], numpy.ndarray):
                        self.results[chainnum][k].resize((len(self.vars[chainnum][k]), iterations/thinning))
                    else:
                        self.results[chainnum][k].resize(iterations/thinning)
                    self.vars[chainnum][k+'_chain'] = self.results[chainnum][k]

        if self.current_iteration[chainnum] < burnin:
            iterthreads = []
            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                iterthreads.append(ModelThread(callback, chainnum, self.current_iteration[chainnum], burnin, [], str(self.eng.outputs['modelcode.cpp'].content), self.scode, self.vars[chainnum], self.data))
                iterthreads[chainnum].start()

            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                iterthreads[chainnum].join()
                logging.info('Finished burnin for chain ' + str(chainnum) + ' :' + str(time.clock() - run_time))

        for chainnum in range(0, int(self.template.EstObjects['nchains'])):
            self.vars[chainnum]['runstate'] = 3
            self.vars[chainnum]['start_iteration'] = self.current_iteration[chainnum] - burnin

        iterthreads = []
        for chainnum in range(0, int(self.template.EstObjects['nchains'])):
            iterthreads.append(ModelThread(callback, chainnum, self.current_iteration[chainnum], total, imputeiters, str(self.eng.outputs['modelcode.cpp'].content), self.scode, self.vars[chainnum], self.data))
            iterthreads[chainnum].start()

        for chainnum in range(0, int(self.template.EstObjects['nchains'])):
            iterthreads[chainnum].join()
            logging.info('Finished iterating for chain ' + str(chainnum) + ' :' + str(time.clock() - run_time))

            for k in self.params[chainnum].keys():
                if k in self.monitor:
                    self.results[chainnum][k] = self.vars[chainnum][k+'_chain']
                else:
                    if isinstance(self.vars[chainnum][k], numpy.ndarray):
                        self.results[chainnum][k].m_n = self.vars[chainnum][str(k)+'_n']
                        self.results[chainnum][k].m_newM = self.vars[chainnum][str(k)+'_oldM']
                        self.results[chainnum][k].m_oldM = self.vars[chainnum][str(k)+'_newM']
                        self.results[chainnum][k].m_newS = self.vars[chainnum][str(k)+'_oldS']
                        self.results[chainnum][k].m_oldS = self.vars[chainnum][str(k)+'_newS']
                    else:
                        self.results[chainnum][k].m_n = self.vars[chainnum][str(k)+'_n']
                        self.results[chainnum][k].m_newM = self.vars[chainnum][str(k)+'_oldM']
                        self.results[chainnum][k].m_oldM = self.vars[chainnum][str(k)+'_newM']
                        self.results[chainnum][k].m_newS = self.vars[chainnum][str(k)+'_oldS']
                        self.results[chainnum][k].m_oldS = self.vars[chainnum][str(k)+'_newS']

                self.params[chainnum][k] = self.vars[chainnum][k]

        self.saveresults()
 
    def genCPP(self, chainnum):
        burnin = int(self.template.EstObjects['burnin'])
        iterations = int(self.template.EstObjects['iterations'])
        thinning = int(self.template.EstObjects['thinning'])
        seed = int(self.template.EstObjects['seed'])
        nchains = int(self.template.EstObjects['nchains'])
        try:
            ncpus = EStat.configuration.getint('EStat', 'standalone_cpus')
        except:
            ncpus = 1

        exprcode = ''

        cppcode = self.data_header_code()
        cppcode += self.param_header_code()
        cppcode += self.adapt_header_code2()


        if hasattr(self.template, 'preccode'):
            cppcode += self.template.preccode

        for k in self.paramnames:
            logging.debug('Param: ' + k)
            if not k in self.customsteps:
                cppcode += '// Update ' + k + '\n'
                cppcode += self.eng.outputs[k + '.xml'].asC(stdtmpl)
                cppcode += '\n'

        if hasattr(self.template, 'postccode'):
            cppcode += self.template.postccode

        from mako import exceptions
        try:
            cppcode = MakoTemplate(cppcode).render(**self.template.objects)
        except:
            cppcode = exceptions.text_error_template().render()

        for d in self.data.keys():
            if isinstance(self.data[d], list) or isinstance(self.data[d], numpy.ndarray):
               cppcode = cppcode.replace('length(' + d + ')', str(len(self.data[d])))
        for p in self.params[0].keys():
            if isinstance(self.params[0][p], list) or isinstance(self.params[0][p], numpy.ndarray):
                cppcode = cppcode.replace('length(' + p + ')', str(len(self.params[0][p])))

        adapt = len(self.needadapt.keys()) > 0

        varnames = {}
        varnames["itcode"] = cppcode        
        varnames["exprcode"] = exprcode
        varnames["params"] = self.params[chainnum]
        varnames["allparams"] = self.params
        varnames["data"] = self.data
        varnames["burnin"] = burnin
        varnames["iterations"] = iterations
        varnames["seed"] = seed
        varnames["thinning"] = thinning
        varnames["needadapt"] = self.needadapt
        varnames["monitorlist"] = self.monitor
        varnames["nchains"] = nchains
        varnames["ncpus"] = ncpus
        
        code = stdtmpl.get_template('standalone.cpp').render(**varnames)
        return code

    def runCPP(self):
        iterations = int(self.template.EstObjects['iterations'])
        thinning = int(self.template.EstObjects['thinning'])
        nchains = int(self.template.EstObjects['nchains'])

        # Write program to disk
        with open(os.path.join(self.eng.directory, 'engine.cpp'), "w") as f:
            f.write(self.eng.outputs['engine.cpp'].content)        # compile program

        args = ['g++', '-std=c++11', '-O3']

        if os.name == 'posix':
            args.append('-o' + os.path.join(self.eng.directory, 'eng'))
        else:
            # args.append('-Wl,--stack,1000000000')
            args.append('-o' + os.path.join(self.eng.directory, 'eng.exe'))

        args.append('-I' + tmplpath)
        args.append('-L' + tmplpath)

        args.append(os.path.join(tmplpath, 'statlib.cpp'))
        args.append(os.path.join(tmplpath, 'rng_simple.cpp'))
        args.append(os.path.join(tmplpath, 'thread_pool.c'))
        args.append(os.path.join(self.eng.directory, 'engine.cpp'))

        args.extend(['-lblas', '-llapack', '-lpthread'])

        p = subprocess.Popen(args)
        p.wait()

        # run program
        exec_name = None
        if os.name == 'posix':
            exec_name = os.path.join(self.eng.directory, 'eng')
        else:
            exec_name = os.path.join(self.eng.directory, 'eng.exe')

        p = subprocess.Popen([exec_name, os.path.join(self.eng.directory, 'output'), os.path.join(self.eng.directory, 'summaries')])
        p.wait()

        for chainnum in range(0, nchains):

            # Allocate space for chains
            for k in self.params[chainnum].keys():
                self.vars[chainnum][k] = self.params[chainnum][k]
                if k in self.monitor:
                    if isinstance(self.vars[chainnum][k], list) or isinstance(self.vars[chainnum][k], numpy.ndarray):
                        self.results[chainnum][k] = numpy.zeros([len(self.vars[chainnum][k]), iterations/thinning])
                    else:
                        self.results[chainnum][k] = numpy.zeros(iterations/thinning)
                else:
                    if isinstance(self.vars[chainnum][k], list) or isinstance(self.vars[chainnum][k], numpy.ndarray):
                        self.results[chainnum][k] = RunningStatVector(len(self.vars[chainnum][k]))
                    else:
                        self.results[chainnum][k] = RunningStat()

            # read chains
            chains = numpy.genfromtxt(os.path.join(self.eng.directory, 'output' + str(chainnum) + '.csv'), names=True, delimiter=',')
            # Copy chains into appropriate structures
            for p in chains.dtype.names:
                if p in self.params[chainnum].keys():
                    self.results[chainnum][p][:] = chains[p]
                else:
                    mat = re.search('([A-Za-z0-9_]*)_([0-9]*)', p)
                    if mat != None:
                        param = mat.group(1)
                        index = int(mat.group(2))
                        if param in self.params[chainnum].keys():
                            self.results[chainnum][param][index][:] = chains[p]

            # read summary stats
            summaries = numpy.genfromtxt(os.path.join(self.eng.directory, 'summaries' + str(chainnum) + '.csv'), names=True, delimiter=',')
            # Copy summaries into appropriate structures
            for p in summaries.dtype.names:
                if p in self.params[chainnum].keys() and isinstance(self.results[chainnum][p], RunningStat):
                    self.results[chainnum][p].m_n = summaries['iteration'][0]
                    self.results[chainnum][p].m_newM = summaries[p][0]
                    self.results[chainnum][p].m_oldM = summaries[p][0]
                    self.results[chainnum][p].m_newS = summaries[p][1] * (self.results[chainnum][p].m_n - 1)
                    self.results[chainnum][p].m_oldS = summaries[p][1] * (self.results[chainnum][p].m_n - 1)
                else:
                    mat = re.search('([A-Za-z0-9_]*)_([0-9]*)', p)
                    if mat != None:
                        param = mat.group(1)
                        index = int(mat.group(2))
                        if param in self.params[chainnum].keys() and isinstance(self.results[chainnum][param], RunningStatVector):
                            self.results[chainnum][param].m_n = summaries['iteration'][0]
                            self.results[chainnum][param].m_newM[index] = summaries[p][0]
                            self.results[chainnum][param].m_oldM[index] = summaries[p][0]
                            self.results[chainnum][param].m_newS[index] = summaries[p][1] * (self.results[chainnum][param].m_n - 1)
                            self.results[chainnum][param].m_oldS[index] = summaries[p][1] * (self.results[chainnum][param].m_n - 1)
        self.saveresults()

    def saveresults(self):
        try:
            displayunmonitored = EStat.configuration.getboolean('EStat', 'displayunmonitored')
        except:
            displayunmonitored = False

        nchains = int(self.template.EstObjects['nchains'])
        niter = int(self.template.EstObjects['iterations'])
        nthin = int(self.template.EstObjects['thinning'])
        mcmc = MCMCChains(nchains, niter, nthin)
        self.eng.outputs['Chains'] = mcmc

        for p in self.params[0].keys():
            if p in self.monitor:
                if self.results[0][p].ndim == 1:
                    chain = numpy.empty((nchains, niter/nthin))
                    for i in range(0, nchains):
                        chain[i, :] = self.results[i][p].copy()
                    mcmc.addchain(p, chain)
                if self.results[0][p].ndim == 2:
                    for j in range(0, self.results[0][p].shape[0]):
                        chain = numpy.empty((nchains, niter/nthin))
                        for i in range(0, int(self.template.EstObjects['nchains'])):
                            chain[i, :] = self.results[i][p][j,:].copy()
                        mcmc.addchain(p + '_' + str(j), chain)

        self.eng.outputs.update(mcmc.summarise(str(self.template.EstObjects['outdata'])))
        results = self.eng.outputs['ModelResults']

        if displayunmonitored == True:
            for p in self.params[0].keys():
                if isinstance(self.results[0][p], RunningStatVector):
                    mean = numpy.zeros(self.results[0][p].length)
                    var = numpy.zeros(self.results[0][p].length)
                    for i in range(0, nchains):
                        mean += self.results[i][p].Mean()
                        var += self.results[i][p].Variance()
                    mean /= nchains
                    var /= nchains
                    B = 0
                    if nchains > 1:
                        for i in range(0, nchains):
                            B += numpy.power(self.results[i][p].Mean() - mean, 2) / (int(self.template.EstObjects['nchains']) - 1)
                        var += (1 + (1/int(self.template.EstObjects['nchains']))) * B

                    for j in range(0, self.results[i][p].length):
                        results.add(p + '_' + str(j), 'mean', mean[j])
                        results.add(p + '_' + str(j), 'sd', numpy.sqrt(var[j]))

                if isinstance(self.results[0][p], RunningStat):
                    mean = 0
                    var = 0
                    for i in range(0, nchains):
                        mean += self.results[i][p].Mean()
                        var += self.results[i][p].Variance()
                    mean /= nchains
                    var /= nchains

                    B = 0
                    if nchains > 1:
                        for i in range(0, nchains):
                            B += numpy.power(self.results[i][p].Mean() - mean, 2) / (nchains - 1)
                        var += (1 + (1/nchains)) * B

                    results.add(p, 'mean', mean)
                    results.add(p, 'std', numpy.sqrt(var))

        for p in self.params[0].keys():
            parents = self.template.objects[p].parents
            if parents != None:
              pos = 0
              for parent in parents:
                if isinstance(parent, DataVector):
                    if p in self.monitor or displayunmonitored == True:
                        origdata = self.dsets[parent.data]
                        for i in range(0, origdata.nobs):
                            results.add(p + '_' + str(i), 'variable', str(parent))

                if isinstance(parent, DataMatrix):
                    origdata = self.dsets[parent.data]
                    for i in range(0, len(parent.orignames)):
                        datavar = origdata.variables[parent.orignames[i]]['data']
                        if parent.orignames[i] in parent.categorical:
                            uniqvals = numpy.unique(datavar)
                            for j in range(1, len(uniqvals)):
                                if origdata.variables[parent.orignames[i]]['label'] == '':
                                    catname = str(uniqvals[j])
                                else:
                                    catname = origdata.valuelabels[origdata.variables[parent.orignames[i]]['label']][uniqvals[j]]

                                if p in self.monitor or displayunmonitored == True:
                                    results.add(p + '_' + str(pos), 'variable', parent.orignames[i] + '_' + catname)
                                pos += 1
                        else:
                            if p in self.monitor or displayunmonitored == True:
                                results.add(p + '_' + str(pos), 'variable', parent.orignames[i])
                            pos += 1

                if isinstance(parent, IDVector):
                    origdata = self.dsets[parent.data]
                    uniqids = numpy.unique(origdata.variables[str(parent)]['data'])

                    if p in self.monitor or displayunmonitored == True:
                        for i in range(0, len(uniqids)):
                            results.add(p + '_' + str(i), 'variable', str(parent))

        parameters = []
        for k in results.content['paramstats'].keys():
            for p in results.content['paramstats'][k].keys():
                if p not in parameters:
                    parameters.append(p)

        stats = results.content['paramstats'].keys()
        paramtable = TabularOutput()
        paramtable.column_headings = ['parameter']
        paramtable.column_headings += stats
        for k in parameters:
            row = list()
            for s in results.content['paramstats'].keys():
                if k in results.content['paramstats'][s].keys():
                    row.append(results.content['paramstats'][s][k])
                else:
                    row.append('')
            paramtable.add_row(k, row)

        self.eng.outputs['ModelParameters'] = paramtable

        if 'dic.cpp' in self.eng.outputs.keys():
            dic = self.dic()
            results.add('model', 'Dbar', dic[0])
            results.add('model', 'D(thetabar)', dic[1])
            results.add('model', 'pD', dic[2])
            results.add('model', 'DIC', dic[3])

            modeltable = TabularOutput()
            modeltable.column_headings = ['Statistic', 'Value']

            for k in results.content['modelstats'].keys():
                modeltable.add_row(k, [results.content['modelstats'][k]])

            self.eng.outputs['ModelFit'] = modeltable

        if self.template.EstObjects['makepred']:
            newdata = {}
            for p in self.params[0].keys():
                parents = self.template.objects[p].parents
                logging.debug('parents of parameter ' + str(p) + ' are ' + str(parents))
                if parents != None:
                  pos = 0
                  for parent in parents:
                    if isinstance(parent, DataVector):
                        origdata = self.dsets[parent.data]
                        if not parent.data in newdata.keys():
                            newdata[parent.data] = DTAFile()
                            newdata[parent.data].nobs = origdata.nobs
                            for var in origdata.variables.keys():
                                newdata[parent.data].addvariable(var, data = origdata.variables[var]['data'])

                        if p in self.monitor:
                            var = numpy.zeros((self.results[0][p].shape[0]))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                var += self.results[chainnum][p].mean(axis = 1)
                            var /= int(self.template.EstObjects['nchains'])
                        else:
                            var = numpy.zeros((self.results[0][p].length))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                var += self.results[chainnum][p].Mean()
                            var /= int(self.template.EstObjects['nchains'])
                        newdata[parent.data].addvariable(p, data = var)

                    if isinstance(parent, DataMatrix):
                        origdata = self.dsets[parent.data]
                        if not 'prediction_' + parent.data in newdata.keys():
                            newdata['prediction_' + parent.data] = DTAFile()
                            newdata['prediction_' + parent.data].nobs = origdata.nobs
                            for var in origdata.variables.keys():
                                newdata['prediction_' + parent.data].addvariable(var, data = origdata.variables[var]['data'])
                            newdata['prediction_' + parent.data].addvariable('pred_full', data = numpy.zeros(newdata['prediction_' + parent.data].nobs))

                        if p in self.monitor:
                            mean = numpy.zeros((self.results[0][p].shape[0]))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                mean += self.results[chainnum][p].mean(axis = 1)
                            mean /= int(self.template.EstObjects['nchains'])

                        else:
                            mean = numpy.zeros((self.results[0][p].length))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                mean += self.results[chainnum][p].Mean()
                            mean /= int(self.template.EstObjects['nchains'])

                        for i in range(0, len(parent.orignames)):
                            datavar = origdata.variables[parent.orignames[i]]['data']
                            if parent.orignames[i] in parent.categorical:
                                uniqvals = numpy.unique(datavar)

                                for j in range(1, len(uniqvals)):
                                    if origdata.variables[parent.orignames[i]]['label'] == '':
                                        catname = str(uniqvals[j])
                                    else:
                                        catname = origdata.valuelabels[origdata.variables[parent.orignames[i]]['label']][uniqvals[j]]

                                    pred = mean[pos] * (datavar == uniqvals[j])
                                    newdata['prediction_' + parent.data].addvariable('pred_' + p + '_' + str(i) + '_' + parent.orignames[i] + '_' + catname, data = pred)
                                    newdata['prediction_' + parent.data].variables['pred_full']['data'] += pred
                                    pos += 1
                            else:
                                pred = mean[pos] * datavar
                                newdata['prediction_' + parent.data].addvariable('pred_' + p + '_' + str(i) + '_' + parent.orignames[i], data = pred)
                                newdata['prediction_' + parent.data].variables['pred_full']['data'] += pred
                                pos += 1

                    if isinstance(parent, IDVector):
                        origdata = self.dsets[parent.data]
                        uniqids = numpy.unique(origdata.variables[str(parent)]['data'])
                        if not str(parent) in newdata.keys():
                            newdata[str(parent)] = DTAFile()
                            newdata[str(parent)].nobs = len(uniqids)
                            newdata[str(parent)].addvariable(str(parent), data = uniqids)

                        if parent.data == 'datafile':
                            ids = self.data[str(parent)]
                        else:
                            ids = self.data[parent.data + '_' + str(parent)]

                        if not 'prediction_' + parent.data in newdata.keys():
                            newdata['prediction_' + parent.data] = DTAFile()
                            newdata['prediction_' + parent.data].nobs = len(ids)
                            for var in origdata.variables.keys():
                                newdata['prediction_' + parent.data].addvariable(var, data = origdata.variables[var]['data'])
                            newdata['prediction_' + parent.data].addvariable('pred_full', data = numpy.zeros(newdata['prediction_' + parent.data].nobs))

                        newdata['prediction_' + parent.data].addvariable(str(parent), data = uniqids[ids])

                        if p in self.monitor:
                            mean = numpy.zeros((self.results[0][p].shape[0]))
                            var = numpy.zeros((self.results[0][p].shape[0]))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                mean += self.results[chainnum][p].mean(axis = 1)
                                var += self.results[chainnum][p].var(axis = 1)
                            mean /= int(self.template.EstObjects['nchains'])
                            var /= int(self.template.EstObjects['nchains'])

                            B = 0
                            if int(self.template.EstObjects['nchains']) > 1:
                                for i in range(0, int(self.template.EstObjects['nchains'])):
                                    B += numpy.power(self.results[chainnum][p].mean(axis = 1) - mean, 2) / (int(self.template.EstObjects['nchains']) - 1)
                            var += (1 + (1/int(self.template.EstObjects['nchains']))) * B

                        else:
                            mean = numpy.zeros((self.results[0][p].length))
                            var = numpy.zeros((self.results[0][p].length))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                mean += self.results[chainnum][p].Mean()
                                var += self.results[chainnum][p].Variance()
                            mean /= int(self.template.EstObjects['nchains'])
                            var /= int(self.template.EstObjects['nchains'])

                            B = 0
                            if int(self.template.EstObjects['nchains']) > 1:
                                for i in range(0, int(self.template.EstObjects['nchains'])):
                                    B += numpy.power(self.results[chainnum][p].Mean() - mean, 2) / (int(self.template.EstObjects['nchains']) - 1)
                            var += (1 + (1/int(self.template.EstObjects['nchains']))) * B

                        newdata[str(parent)].addvariable(p, data = mean)
                        newdata[str(parent)].addvariable(p+'_sd', data = numpy.sqrt(var))
                        newdata['prediction_' + parent.data].addvariable('pred_' + p, data = mean[ids])
                        newdata['prediction_' + parent.data].variables['pred_full']['data'] += mean[ids]

                    if isinstance(parent, IDMatrix):
                        if not str(IDMatrix) in newdata.keys():
                            origdata = self.dsets[parent.data]
                            a = []
                            for i in range(0, parent.ncols()):
                                a += list(origdata.variables[parent[i]]['data'].compressed())
                            IDS = numpy.unique(a)
                            newdata[str(parent)] = DTAFile()
                            newdata[str(parent)].nobs = len(IDS)
                            newdata[str(parent)].addvariable(str(parent), data = numpy.unique(IDS))
                        if p in self.monitor:
                            var = numpy.zeros((self.results[0][p].shape[0]))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                var += self.results[chainnum][p].mean(axis = 1)
                            var /= int(self.template.EstObjects['nchains'])
                            newdata[str(parent)].addvariable(p, data = var)
                        else:
                            var = numpy.zeros((self.results[0][p].length))
                            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                                var += self.results[chainnum][p].Mean()
                            var /= int(self.template.EstObjects['nchains'])
                            newdata[str(parent)].addvariable(p, data = var)

            for k in newdata.keys():
                logging.debug('creating new dataset: ' + k)
                self.eng.outputs[k] = newdata[k]

        haspred = False
        for p in self.params[0].keys():        
            if p.startswith('pred'):
                 haspred = True
                 break

        if haspred:
            origdata = self.dsets['datafile']
            dta = DTAFile()
            dta.nobs = origdata.nobs
            # First copy the original variables
            for var in origdata.variables.keys():
                dta.addvariable(var, data = origdata.variables[var]['data'])
            # Now add the predicted variables
            for p in self.params[0].keys():
                if p.startswith('pred'):
                     dta.addvariable(p, data = self.results[chainnum][p].Mean());

    def dic(self):
        diccode = self.eng.outputs['dic.cpp'].content

        meanvars = {}
        for p in self.params[0].keys():
            if p in self.monitor:
                if len(self.results[0][p].shape) > 1:
                    meanvars[p] = numpy.zeros(self.results[0][p].shape[0], dtype=float)
                    for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                        meanvars[p] += self.results[chainnum][p].mean(axis=1)
                else:
                    meanvars[p] = 0.0
                    for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                        meanvars[p] += float(self.results[chainnum][p].mean())
            else:
                if isinstance(self.results[0][p], RunningStatVector):
                    meanvars[p] = numpy.zeros(self.results[0][p].length, dtype=float)
                    for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                        meanvars[p] += numpy.array(self.results[chainnum][p].Mean())
                else:
                    meanvars[p] = 0.0
                    for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                        meanvars[p] += float(self.results[chainnum][p].Mean())

            meanvars[p] /= int(self.template.EstObjects['nchains'])

        meandeviance = 0
        if 'deviance' in self.monitor:
            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                meandeviance += numpy.mean(self.results[chainnum]['deviance'])
        else: 
            for chainnum in range(0, int(self.template.EstObjects['nchains'])):
                meandeviance += float(self.results[chainnum]['deviance'].Mean())
        meandeviance /= int(self.template.EstObjects['nchains'])

        # Deviance is calculated over all chains
        meanvars['numchains'] = 1
        meanvars['chainnum'] = 0

        self.runcode(str(diccode), meanvars, self.data)

        deviancemean = meanvars['deviance']
        pD = meandeviance - deviancemean
        DIC = meandeviance + pD

        return (meandeviance, deviancemean, pD, DIC)
        
